/* translatable strings for gretl addons */

const char *gig_translations[] = {
    N_("gig"),
    N_("APARCH"),
    N_("AR lags"),
    N_("ARCH"),
    N_("Constant"),
    N_("Covariance estimator"),
    N_("Density"),
    N_("Dependent Variable"),
    N_("Distribution"),
    N_("EGARCH"),
    N_("GARCH"),
    N_("GED"),
    N_("GJR"),
    N_("Hessian"),
    N_("Mean regressors"),
    N_("Model type"),
    N_("NARCH"),
    N_("Normal"),
    N_("OPG"),
    N_("Plot type"),
    N_("Sandwich"),
    N_("Skewed GED"),
    N_("Skewed t"),
    N_("TARCH"),
    N_("Taylor/Schwert GARCH"),
    N_("Time series"),
    N_("Variance regressors"),
    N_("Verbosity"),
    N_("t")
};

const char *ivpanel_translations[] = {
    N_("Between model"),
    N_("Fixed effects"),
    N_("G2SLS"),
    N_("dependent variable"),
    N_("instruments"),
    N_("regressors")
};

const char *SVAR_translations[] = {
    N_("SVAR"),
    N_("AB-model"),
    N_("Bootstrap alpha"),
    N_("Bootstrap replications"),
    N_("C-model"),
    N_("Constant"),
    N_("Exogenous regressors"),
    N_("IRF by shock"),
    N_("IRF by var"),
    N_("KPSW"),
    N_("Lags"),
    N_("Model type"),
    N_("Plot type"),
    N_("Time trend"),
    N_("VAR variables"),
    N_("horizon"),
    N_("plain"),
    N_("template matrix")
};

